import asyncio
import logging
from os import mkdir, path
from typing import Optional, TextIO

import asyncclick as click
from hypercorn.config import Config as HConfig
from hypercorn.asyncio import serve
from xdg import XDG_CONFIG_HOME, XDG_DATA_HOME

from escarpolette import create_app
from escarpolette.settings import Default, Config


DEFAULT_CONFIG_FOLDER = path.join(XDG_CONFIG_HOME, "escarpolette")
DEFAULT_DATA_FOLDER = path.join(XDG_DATA_HOME, "escarpolette")


@click.command()
@click.option(
    "--config", "config_file", help="path to the config file", type=click.File("w+")
)
@click.option(
    "--host",
    default=Default.HOST,
    help="address to bind to listen for connections",
    type=str,
)
@click.option(
    "--port",
    default=Default.PORT,
    help="port to bind to listen for connections",
    type=int,
)
@click.option(
    "--dev",
    default=False,
    help="enable code autoreload and debug logs",
    type=bool,
    is_flag=True,
)
async def run(config_file: Optional[TextIO], host: str, port: int, dev: bool) -> None:
    """Run the application.

    Create default folders for config and data and the default config file if
    no one is given, create the fastapi app and start the server.

    TODO:
        * setup loging
        * use a FastAPI's "start" event to setup the app, and switch back to uvicorn?
    """
    if not path.exists(DEFAULT_CONFIG_FOLDER):
        mkdir(DEFAULT_CONFIG_FOLDER)

    if not path.exists(DEFAULT_DATA_FOLDER):
        mkdir(DEFAULT_DATA_FOLDER)

    ######################
    # read configuration #
    ######################

    if config_file is None:
        config_file = open(
            path.join(XDG_CONFIG_HOME, "escarpolette", "escarpolette.conf"), "w+"
        )

    config = Config(config_file)
    config_file.close()

    ###############
    # set logging #
    ###############

    level = logging.INFO
    if dev:
        level = logging.DEBUG

    logging.basicConfig(
        level=level,
        format="%(asctime)s - %(levelname)s - %(name)s:%(lineno)s - %(message)s",
    )

    ###################
    # create ASGI app #
    ###################
    app = await create_app(config)

    ####################
    # start web server #
    ####################

    host = host or config.HOST
    port = port or config.PORT

    server_config = HConfig()
    server_config.bind = [f"{host}:{port}"]
    loop = asyncio.get_event_loop()
    future = loop.create_task(serve(app, server_config))
    await future


run(_anyio_backend="asyncio")



from pydantic import HttpUrl, Field
from pydantic.main import BaseModel


class BaseItemSchema(BaseModel):
    url: HttpUrl = Field(..., example="https://www.youtube.com/watch?v=bpA6fAz_r04")


class ItemSchemaIn(BaseItemSchema):
    pass


class ItemSchemaOut(BaseItemSchema):
    class Config:
        orm_mode = True
        extra = "ignore"

    artist: str = Field(..., example="Vic Dibitetto")
    duration: int = Field(..., example=94)
    id: int
    played: bool
    title: str = Field(..., example="Anybody want cawfee?!")



"""This module is only  intended to be use with uvicorn "import path" mode."""

from escarpolette import create_app
from escarpolette.settings import Config


config = Config.current_config
if config is None:
    raise RuntimeError("The config was not initialised")
else:
    app = create_app(config)



from fastapi import Depends
from sqlalchemy.orm import Session

from escarpolette import db
from escarpolette.models import Playlist


def get_db():
    with db.get_db() as db_session:
        yield db_session


def get_current_playlist(db: Session = Depends(get_db)):
    playlist = Playlist.get_current_playlist(db)
    return playlist



import asyncio
import json
import logging
from enum import Enum
from subprocess import Popen
from typing import Any, Dict, Optional, Tuple

from pydantic import BaseModel, ValidationError
from pydantic.fields import Field

from escarpolette.models import Playlist
from escarpolette.settings import Config


logger = logging.getLogger(__name__)


class MpvEvent(BaseModel):
    name: str = Field(..., alias="event")


class MpvResponse(BaseModel):
    data: Any
    error: str
    request_id: int


class State(str, Enum):
    PLAYING = "PLAYING"
    PAUSED = "PAUSED"
    STOPPED = "STOPPED"


class PlayerCommandError(ValueError):
    pass


class Player:
    """Control mpv using IPC."""

    _command_id = 0
    _mpv_ipc_socket = "/tmp/mpv-socket"
    _mpv: Optional[Popen] = None
    _mpv_reader: asyncio.StreamReader
    _mpv_writer: asyncio.StreamWriter
    _state = State.STOPPED

    async def init_app(self, config: Config) -> None:
        self._mpv_ipc_socket = config.MPV_IPC_SOCKET or self._mpv_ipc_socket
        self._mpv = Popen(
            [
                "mpv",
                "--idle",
                "--no-terminal",
                f"--input-ipc-server={self._mpv_ipc_socket}",
            ]
        )

        # Let MPV starts
        await asyncio.sleep(2)

        loop = asyncio.get_event_loop()
        loop.create_task(self._listen_events())
        self._mpv_reader, self._mpv_writer = await self._get_mpv_connection()

    def shutdown(self) -> None:
        if self._mpv_writer is not None:
            self._mpv_writer.close()

        if self._mpv is not None:
            # TODO: find why MPV does not respond to a SIGTERM signal
            self._mpv.kill()

    async def add_item(self, url: str) -> None:
        """Add a new item to the playlist.

        If the player was stopped, play the music.
        """
        if self._state == State.STOPPED:
            await self._send_command("loadfile", url, "append-play")
            self._state = State.PLAYING
        else:
            await self._send_command("loadfile", url, "append")

    async def get_current_item_title(self) -> Optional[str]:
        """Get the current playing item's title."""
        response = await self._send_command("get_property", "metadata")
        if not response:
            return None

        return response.data["title"]

    async def get_current_item_position(self) -> Optional[int]:
        response = await self._send_command("get_property", "playback-time")
        if not response:
            return None

        return response.data

    def get_state(self) -> State:
        return self._state

    async def play(self) -> None:
        """Play the current playlist."""
        if self._state == State.PLAYING:
            return
        elif self._state == State.STOPPED:
            raise PlayerCommandError("The player is stopped. Add an item to play it.")
        else:
            await self._send_command("cycle", "pause")

        return None

    async def pause(self) -> None:
        """Pause the current playlist."""
        if self._state == State.PAUSED:
            return
        elif self._state == State.STOPPED:
            raise PlayerCommandError("The player is stopped.")

        await self._send_command("cycle", "pause")
        return None

    def _get_command_id(self) -> int:
        self._command_id += 1
        return self._command_id

    async def _get_mpv_connection(
        self,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        logger.info("Connecting to MVP on %s", self._mpv_ipc_socket)
        reader, writer = await asyncio.open_unix_connection(self._mpv_ipc_socket)
        logger.info("Connected to MVP")
        return reader, writer

    @staticmethod
    async def _get_mpv_message(reader: asyncio.StreamReader) -> Optional[Dict]:
        data = await reader.readuntil(b"\n")
        logger.debug("Received data from MVP: %s", data)
        try:
            message = json.loads(data)
        except json.JSONDecodeError as e:
            logger.debug("Cannot decode MVP data: %s", e)
            return None

        logger.debug("Received MVP message %s", message)
        return message

    async def _listen_events(self):
        """Listen for events from MVP.

        Open a connection to MVP, listen for events and update the playlist
        accordingly.
        """
        reader, _ = await self._get_mpv_connection()

        while True:
            message = await self._get_mpv_message(reader)

            try:
                event = MpvEvent(**message)
            except ValidationError as e:
                logger.debug("Received unknown message from MPV: %s", e)
                continue

            if event.name == "end-file":
                logger.info("Track ended")
                Playlist.item_ended()
            elif event.name == "idle":
                logger.info("Player stopped")
                self._state = State.STOPPED
            elif event.name == "pause":
                logger.info("Player paused")
                self._state = State.PAUSED
            elif event.name == "start-file":
                logger.info("Player playing")
                self._state = State.PLAYING
            elif event.name == "unpause":
                logger.info("Player playing")
                self._state = State.PLAYING
            else:
                logger.debug("Unknown MPV event %s", event.name)

    async def _send_command(self, *command: str) -> Optional[MpvResponse]:
        """Send a command to MPV and return the response."""
        command_id = self._get_command_id()
        msg = {"command": command, "request_id": command_id}
        logger.debug("Sending MPV command %s", msg)

        data = json.dumps(msg).encode("utf8") + b"\n"
        self._mpv_writer.write(data)
        await self._mpv_writer.drain()

        return await self._read_response(command_id)

    async def _read_response(self, command_id: int) -> Optional[MpvResponse]:
        while True:
            message = await self._get_mpv_message(self._mpv_reader)
            if not message:
                return None

            if "event" in message:
                logger.debug("Skipping event")
                continue

            try:
                response = MpvResponse(**message)
            except ValidationError:
                logger.warning("Received unknown message from MPV: %s", message)
                continue

            if response.request_id == command_id:
                return response


_current_player = Player()


def get_player():
    return _current_player



from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

from escarpolette.db import Base, get_db
from escarpolette.models.mixin import BaseModelMixin


class Playlist(BaseModelMixin, Base):
    __tablename__ = "playlists"

    items = relationship(
        "Item",
        back_populates="playlist",
        order_by="Item.position",
        collection_class=ordering_list("position"),
    )

    @classmethod
    def get_current_playlist(cls, db_session):
        playlist = db_session.query(cls).order_by(Playlist.created_at.desc()).first()
        return playlist

    @classmethod
    def item_ended(cls):
        """Mark current playing item as played."""
        with get_db() as db_session:
            playlist = cls.get_current_playlist(db_session)

            played = True
            for item in playlist.items:
                if not item.played:
                    item.played = True
                    break

            db_session.commit()



from escarpolette.schemas.item import ItemSchemaIn, ItemSchemaOut
from escarpolette.schemas.player import PlayerSchema
from escarpolette.schemas.playlist import PlaylistSchemaOut

__all__ = ["ItemSchemaIn", "ItemSchemaOut", "PlaylistSchemaOut"]



import logging

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from .dependencies import get_db, get_current_playlist
from escarpolette.login import get_current_user, User
from escarpolette.models import Item, Playlist
from escarpolette.player import get_player, Player, State
from escarpolette.schemas.item import ItemSchemaIn, ItemSchemaOut
from escarpolette.schemas.playlist import PlayingItem, PlaylistSchemaOut
from escarpolette.tools import get_content_metadata
from escarpolette.rules import rules


logger = logging.getLogger(__name__)
router = APIRouter()

ERROR_CONFLICT_MSG = "This item is already waiting to be played"


@router.get("/", response_model=PlaylistSchemaOut)
async def get(
    current_playlist: Playlist = Depends(get_current_playlist),
    player: Player = Depends(get_player),
) -> PlaylistSchemaOut:
    playlist = PlaylistSchemaOut()
    is_stopped = player.get_state() == State.STOPPED

    for item in current_playlist.items:
        playlist.items.append(ItemSchemaOut.from_orm(item))
        if not is_stopped and not item.played and playlist.playing is None:
            playlist.playing = PlayingItem(
                id=item.id, duration=await player.get_current_item_position()
            )
    return playlist


@router.post(
    "/",
    status_code=201,
    response_model=ItemSchemaOut,
    responses={409: {"description": ERROR_CONFLICT_MSG}},
)
async def post(
    data: ItemSchemaIn,
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
    playlist: Playlist = Depends(get_current_playlist),
    player: Player = Depends(get_player),
) -> Item:
    metadata = get_content_metadata(data.url)
    item = Item(user_id=current_user.id, **metadata)

    if not rules.can_add_item(current_user, item, db):
        raise TooManyRequests

    playlist.items.append(item)
    db.add(playlist)

    try:
        db.flush()
    except IntegrityError as e:
        logger.debug("Integrity error while trying to add a new item: %s", e)
        raise HTTPException(status_code=409, detail=ERROR_CONFLICT_MSG)

    await player.add_item(item.url)
    db.commit()

    return item



from sqlalchemy import text, Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from escarpolette.db import Base
from escarpolette.models.mixin import BaseModelMixin


class Item(BaseModelMixin, Base):
    __tablename__ = "items"

    artist = Column(String(255))
    duration = Column(Integer)
    played = Column(Boolean, default=False, server_default=text("FALSE"))
    position = Column(Integer)
    title = Column(String(255))
    url = Column(String(255))
    user_id = Column(String(36), index=True, nullable=False)
    playlist_id = Column(
        Integer,
        ForeignKey("playlists.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )

    playlist = relationship("Playlist", back_populates="items")



from datetime import timedelta
from configparser import ConfigParser
from typing import TextIO
from uuid import uuid4

from xdg import XDG_DATA_HOME


class Default:
    # Server
    HOST = "127.0.0.1"
    PORT = 8000

    # Database
    # SQLALCHEMY_TRACK_MODIFICATIONS = False
    DATABASE_URI = f"sqlite:///{XDG_DATA_HOME}/escarpolette/db.sqlite"

    # Authentication
    REMEMBER_COOKIE_DURATION = timedelta(days=390)  # ~13 months


class Config(Default):
    current_config = None

    def __init__(self, file: TextIO):
        """Read the config from a given file.

        If some values are missing, apply defaults. In any case, write the
        resulting config to the same file.
        """
        config = ConfigParser()
        config.read_file(file)

        if "DATABASE" not in config:
            config["DATABASE"] = {}

        if "MPV" not in config:
            config["MPV"] = {}

        if "SECURITY" not in config:
            config["SECURITY"] = {}

        if "SERVER" not in config:
            config["SERVER"] = {}

        self.HOST = config["SERVER"].get("HOST", self.HOST)
        self.PORT = config["SERVER"].getint("PORT", fallback=self.PORT)
        self.DATABASE_URI = config["DATABASE"].get("URI", self.DATABASE_URI)
        self.MPV_IPC_SOCKET = config["MPV"].get("IPC_SOCKET")
        self.SECRET_KEY = config["SECURITY"].get("SECRET_KEY", str(uuid4()))

        config["DATABASE"] = {"URI": self.DATABASE_URI}

        if self.MPV_IPC_SOCKET is not None:
            config["MVP"] = {"IPC_SOCKET": self.MPV_IPC_SOCKET}

        config["SECURTIY"] = {"SECRET_KEY": self.SECRET_KEY}
        config["SERVER"] = {"HOST": self.HOST, "PORT": str(self.PORT)}

        config.write(file)

        # save a singleton
        Config.current_config = self


def get_current_config():
    return Config.current_config



from contextlib import contextmanager
from importlib import resources

from sqlalchemy import Column, create_engine, String
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.orm import sessionmaker, Session

from escarpolette.settings import Config

Base: DeclarativeMeta = declarative_base()
SessionLocal = sessionmaker(autocommit=False, autoflush=False)


class Version(Base):
    __tablename__ = "version"

    version = Column(String, primary_key=True)


def init_app(config: Config):
    engine = create_engine(
        config.DATABASE_URI, connect_args={"check_same_thread": False}
    )
    SessionLocal.configure(bind=engine)

    apply_migrations(engine)


def apply_migrations(engine):
    Base.metadata.create_all(bind=engine, tables=[Version.__table__], checkfirst=True)

    db: Session = SessionLocal()
    current_version = db.query(Version).first()
    if current_version is None:
        current_version = Version(version="v000")

    migrations = resources.contents("escarpolette.migrations")
    for migration in sorted(migrations):
        if not migration.endswith(".sql"):
            continue

        migration_version, _ = migration.split(".")
        if migration_version <= current_version.version:
            continue

        sql = resources.read_text("escarpolette.migrations", migration)
        queries = sql.split(";")
        for query in queries:
            db.execute(query)

        current_version.version, _ = migration.split(".")
        db.add(current_version)
        db.commit()


@contextmanager
def get_db():
    db: Session = SessionLocal()
    db.execute("PRAGMA foreign_keys = ON")
    try:
        yield db
    finally:
        db.close()



import logging

from fastapi import APIRouter, Depends, HTTPException

from escarpolette.player import State, get_player, Player
from escarpolette.schemas import PlayerSchema

logger = logging.getLogger(__name__)
router = APIRouter()


@router.get("/", response_model=PlayerSchema)
async def get(player: Player = Depends(get_player)) -> PlayerSchema:
    logger.debug("GET player state")
    data = PlayerSchema(state=player.get_state().name)
    return data


@router.patch("/", responses={400: {}})
async def patch(data: PlayerSchema, player: Player = Depends(get_player)) -> None:
    if data.state == State.PLAYING:
        await player.play()
    elif data.state == State.PAUSED:
        await player.pause()
    else:
        raise HTTPException(
            status_code=400, detail=f"The state {data.state} cannot be set"
        )



from .items import router as item_router
from .player import router as player_router


def init_app(app):
    app.include_router(item_router, prefix="/item", tags=["items"])
    app.include_router(player_router, prefix="/player", tags=["player"])



from dataclasses import dataclass

from sqlalchemy import func
from sqlalchemy.orm import Session

from escarpolette.login import User
from escarpolette.models.item import Item


@dataclass
class RulesConfig:
    MAX_ITEM_NUMBER: int = 3
    MAX_ITEM_LENGTH: int = 600


class Rules:
    def __init__(self, config: RulesConfig) -> None:
        self.config = config

    def can_add_item(self, user: User, item: Item, db: Session) -> bool:
        if item.duration > self.config.MAX_ITEM_LENGTH:
            return False

        count = (
            db.query(Item)
            .filter(Item.created_at >= func.now(), Item.user_id == user.id)
            .count()
        )

        return count < self.config.MAX_ITEM_NUMBER

    def can_remove_item(self, user: User) -> bool:
        return False

    def can_toogle_pause(self, user: User) -> bool:
        return False

    def can_vote(self, user: User, item: Item) -> bool:
        return False



from fastapi import FastAPI

from escarpolette.player import Player
from escarpolette.settings import Config


def init_app(app: FastAPI, config: Config):
    # CORS(app)
    # Migrate(app, db)
    pass



from uuid import uuid4

import jwt
from fastapi import Cookie, Depends
from jwt.exceptions import DecodeError
from pydantic import BaseModel
from starlette.responses import Response

from escarpolette.settings import Config, get_current_config


MAX_AGE = 13 * 30 * 24 * 60 * 60  # 13 months
ALGORITHM = "HS256"


class User(BaseModel):
    id: str


def get_current_user(
    response: Response,
    token: str = Cookie(None),
    config: Config = Depends(get_current_config),
):
    if not token:
        current_user = User(id=str(uuid4()))
    else:
        try:
            claim = jwt.decode(token, config.SECRET_KEY, algorithms=[ALGORITHM])
        except DecodeError:
            current_user = User(id=str(uuid4()))
        else:
            current_user = User(id=claim["user_id"])

    payload = {"user_id": current_user.id}
    sign = jwt.encode(payload, config.SECRET_KEY, algorithm=ALGORITHM).decode("utf-8")
    response.set_cookie("token", sign, MAX_AGE)

    return current_user



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['escarpolette',
 'escarpolette.admin',
 'escarpolette.migrations',
 'escarpolette.models',
 'escarpolette.routers',
 'escarpolette.rules',
 'escarpolette.schemas']

package_data = \
{'': ['*']}

install_requires = \
['asyncclick>=7.0.9,<8.0.0',
 'fastapi>=0.47.1,<0.48.0',
 'hypercorn>=0.9.0,<0.10.0',
 'pydantic>=1.3,<2.0',
 'pyjwt>=1.7.1,<2.0.0',
 'sqlalchemy>=1.2,<2.0',
 'xdg>=4.0,<5.0',
 'youtube-dl']

entry_points = \
{'console_scripts': ['escarpolette = escarpolette:__main__']}

setup_kwargs = {
    'name': 'escarpolette',
    'version': '0.6.0',
    'description': 'Manage your musical playlist with your friends without starting a war.',
    'long_description': "# Escarpolette\n\nThis project provides a server and clients to manage your music playlist when you are hosting a party.\n\nIt supports many sites thanks to the awesome project [youtube-dl](https://rg3.github.io/youtube-dl/).\n\n## Features\n\nServer:\n* add items (and play them!)\n* get playlist's items\n* runs on Android! (see [instructions](#Android))\n\nWeb client:\n* there is currently no web client :(\n\n## Dependencies\n\n* Python 3.7\n* the player [mpv](https://mpv.io)\n\nThey should be available for most of the plateforms.\n\n\n## Installation\n\n```Shell\npip install escarpolette\n```\n\n### Android\n\nYou will need [Termux](https://termux.com/).\nThen inside Termux you can install it with:\n\n```Shell\n# dependencies\npkg install python python-dev clang mpv\n# escarpolette\npip install escarpolette\n```\n\nNote that while the project can run without wake-lock, acquiring it improve the performance (with a battery trade off).\n\n## Usage\n\n```Shell\nescarpolette [--config config.cfg] [--host host] [--port port] [--help]\n```\n\nThe default configuration should be good for all the usages.\n\n## Dev\n\nYou will need [Poetry](https://poetry.eustace.io/) to manage the dependencies.\n\nClone the repo and then type `poetry install`.\nYou can run the app with `poetry run python -m escarpolette`.\n\n## Todo\n\n* server\n    * votes\n    * bonjour / mDNS\n    * prevent adding youtube / soundcloud playlists\n    * restrictions by users\n    * configuration of those restrictions by an admin\n* web client\n    * show playing status\n    * votes\n    * configure restrictions:\n        * max video added per user\n        * max video length\n    * admin access:\n        * configure restrictions\n        * no restrictions for him\n        * force video order\n\nMaybe one day?\n* android client\n* iOS client\n",
    'author': 'Alexandre Morignot',
    'author_email': 'erdnaxeli@cervoi.se',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/erdnaxeli/escarpolette',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from sqlalchemy import Column, Integer, DateTime, text


class BaseModelMixin:
    id = Column(Integer, primary_key=True)
    created_at = Column(
        DateTime,
        default=datetime.now,
        server_default=text("datetime()"),
        index=True,
        nullable=False,
    )



import logging

from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from sqlalchemy.orm import Session

from escarpolette import db, routers
from escarpolette.models import Playlist
from escarpolette.settings import Config
from escarpolette.player import get_player

request_logger = logging.getLogger("escarpolette")


def create_new_playlist(db: Session):
    playlist = Playlist()
    db.add(playlist)
    db.commit()


async def create_app(config: Config):
    app = FastAPI(
        title="Escarpolette",
        version="0.1",
        description="Manage your party's playlist without friction",
    )

    routers.init_app(app)
    db.init_app(config)
    await get_player().init_app(config)

    @app.on_event("shutdown")
    def shutdown():
        get_player().shutdown()

    with db.get_db() as db_session:
        create_new_playlist(db_session)

    app.add_middleware(
        CORSMiddleware, allow_credentials=True, allow_methods=["*"], allow_origins=["*"]
    )

    @app.middleware("http")
    async def log_request(request: Request, call_next):
        request_logger.info("%s %s", request.method, request.url.path)
        response = await call_next(request)
        return response

    return app



from typing import List, Optional

from pydantic import BaseModel, Field

from escarpolette.schemas.item import ItemSchemaOut


class PlayingItem(BaseModel):
    id: int = Field(..., description="The id of the item currently playing")
    duration: int = Field(
        ..., description="The part, in seconds, of the item already played."
    )


class PlaylistSchemaOut(BaseModel):
    items: List[ItemSchemaOut] = []
    playing: PlayingItem = Field(
        None, description="Info about the current playing item"
    )
